/*  text2bin.c - source code for generating binary image
 *               files from text containing ASCII and
 *               PC compatible extended characters
 *
 *  Fax Compression Sample Code
 *
 *  Copyright (C) 2000  Jack Klein
 *                      JK Technology
 *                      Macmillan Computer Publishing
 *
 *  This program is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General
 *  Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will
 *  be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General
 *  Public License along with this program; if not, write
 *  to the Free Software Foundation, Inc., 675 Mass Ave,
 *  Cambridge, MA 02139, USA.
 *
 *  Jack Klein may be contacted by email at:
 *     The_C_Guru@yahoo.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fax.h"

/* definitions, based on fax.h  */
#define OCTETS_PER_CHAR   2
#define CHARS_PER_ROW     (OCTETS_PER_ROW / OCTETS_PER_CHAR)
#define SCAN_ROWS_PER_CHAR  16
#define LINES_PER_PAGE (MAXIMUM_ROWS / SCAN_ROWS_PER_CHAR)

/*  The character look-up array contains an array of 16
 *  unsigned short integers for each character
 *
 *  Each unsigned short int contains the bit pattern for
 *  one scan row of the character, using the least
 *  significant 16 bits of the unsigned short, even if
 *  unsigned short holds more than 16 bits
 *
 *  The bit patterns are defined as they would be scanned
 *  by a typical scanner, with 0 bits for black where the
 *  character appears, and 1 bits for the background paper
 *
 *  Each character occupies a 16 pixel width (2 mm per
 *  character, approximately 12.5 characters per inch
 *
 *  For standard fax resolution each horizontal row
 *  generated should be output once, for high resolution
 *  each row should be output twice
 */
static const unsigned short CharacterSet [256] [16] =
{
  /* character 00 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 01 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0x3FFC,  /* 0011111111111100  ..************.. */
    0x33CC,  /* 0011001111001100  ..**..****..**.. */
    0x3FFC,  /* 0011111111111100  ..************.. */
    0x3FFC,  /* 0011111111111100  ..************.. */
    0x300C,  /* 0011000000001100  ..**........**.. */
    0x3C3C,  /* 0011110000111100  ..****....****.. */
    0x3FFC,  /* 0011111111111100  ..************.. */
    0x3FFC,  /* 0011111111111100  ..************.. */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 02 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x3033,  /* 0011000000110011  ..**......**..** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 03 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0303,  /* 0000001100000011  ......**......** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 04 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 05 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0303,  /* 0000001100000011  ......**......** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 06 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 07 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 08 =   */
  {
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x03C0,  /* 0000001111000000  ......****...... */
    0x0FF0,  /* 0000111111110000  ....********.... */
    0x0FF0,  /* 0000111111110000  ....********.... */
    0x0FF0,  /* 0000111111110000  ....********.... */
    0x03C0,  /* 0000001111000000  ......****...... */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
  },
  /* character 09 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 0A =   */
  {
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x03C0,  /* 0000001111000000  ......****...... */
    0x0FF0,  /* 0000111111110000  ....********.... */
    0x3C3C,  /* 0011110000111100  ..****....****.. */
    0x3C3C,  /* 0011110000111100  ..****....****.. */
    0x3C3C,  /* 0011110000111100  ..****....****.. */
    0x0FF0,  /* 0000111111110000  ....********.... */
    0x03C0,  /* 0000001111000000  ......****...... */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
  },
  /* character 0B =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 0C =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 0D =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFC33,  /* 1111110000110011  ******....**..** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 0E =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 0F =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0xC003,  /* 1100000000000011  **............** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC003,  /* 1100000000000011  **............** */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 10 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x3FFF,  /* 0011111111111111  ..************** */
    0x03FF,  /* 0000001111111111  ......********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x000F,  /* 0000000000001111  ............**** */
    0x0003,  /* 0000000000000011  ..............** */
    0x000F,  /* 0000000000001111  ............**** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x03FF,  /* 0000001111111111  ......********** */
    0x3FFF,  /* 0011111111111111  ..************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 11 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFF3,  /* 1111111111110011  ************..** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xC003,  /* 1100000000000011  **............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC003,  /* 1100000000000011  **............** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFFF3,  /* 1111111111110011  ************..** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 12 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 13 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 14 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC000,  /* 1100000000000000  **.............. */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0xC030,  /* 1100000000110000  **........**.... */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 15 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 16 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 17 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 18 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 19 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1A =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF03,  /* 1111111100000011  ********......** */
    0x0000,  /* 0000000000000000  ................ */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1B =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1C =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1D =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF3CF,  /* 1111001111001111  ****..****..**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0000,  /* 0000000000000000  ................ */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF3CF,  /* 1111001111001111  ****..****..**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1E =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 1F =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 20 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 21 = ! */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 22 = " */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xFCCF,  /* 1111110011001111  ******..**..**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 23 = # */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 24 = $ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC03F,  /* 1100000000111111  **........****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 25 = % */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 26 = & */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xC003,  /* 1100000000000011  **............** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 27 = ' */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 28 = ( */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 29 = ) */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2A = * */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0x0003,  /* 0000000000000011  ..............** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2B = + */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2C = , */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2D = - */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2E = . */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 2F = / */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFF3,  /* 1111111111110011  ************..** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x3FFF,  /* 0011111111111111  ..************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 30 = 0 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 31 = 1 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 32 = 2 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 33 = 3 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 34 = 4 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 35 = 5 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x000F,  /* 0000000000001111  ............**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 36 = 6 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x000F,  /* 0000000000001111  ............**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 37 = 7 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 38 = 8 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 39 = 9 */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3A = : */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3B = ; */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3C = < */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3D = = */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3E = > */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 3F = ? */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 40 = @ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 41 = A */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 42 = B */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x000F,  /* 0000000000001111  ............**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x000F,  /* 0000000000001111  ............**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 43 = C */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0FF3,  /* 0000111111110011  ....********..** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FF3,  /* 0000111111110011  ....********..** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 44 = D */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 45 = E */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3CF,  /* 1100001111001111  **....****..**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3CF,  /* 1100001111001111  **....****..**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 46 = F */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3CF,  /* 1100001111001111  **....****..**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3CF,  /* 1100001111001111  **....****..**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 47 = G */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 48 = H */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 49 = I */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4A = J */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4B = K */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4C = L */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4D = M */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0303,  /* 0000001100000011  ......**......** */
    0x0303,  /* 0000001100000011  ......**......** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4E = N */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 4F = O */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 50 = P */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x000F,  /* 0000000000001111  ............**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 51 = Q */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 52 = R */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x000F,  /* 0000000000001111  ............**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 53 = S */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 54 = T */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xCC33,  /* 1100110000110011  **..**....**..** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 55 = U */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 56 = V */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 57 = W */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0303,  /* 0000001100000011  ......**......** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 58 = X */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 59 = Y */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5A = Z */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x3FC3,  /* 0011111111000011  ..********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FF3,  /* 0000111111110011  ....********..** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5B = [ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5C = \ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x3FFF,  /* 0011111111111111  ..************** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFF3,  /* 1111111111110011  ************..** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5D = ] */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5E = ^ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 5F = _ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 60 = ` */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 61 = a */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 62 = b */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x03FF,  /* 0000001111111111  ......********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x000F,  /* 0000000000001111  ............**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 63 = c */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 64 = d */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 65 = e */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 66 = f */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0x000F,  /* 0000000000001111  ............**** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 67 = g */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 68 = h */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x03FF,  /* 0000001111111111  ......********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 69 = i */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6A = j */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6B = k */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x03FF,  /* 0000001111111111  ......********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6C = l */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6D = m */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6E = n */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 6F = o */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 70 = p */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 71 = q */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 72 = r */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 73 = s */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 74 = t */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0x000F,  /* 0000000000001111  ............**** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 75 = u */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 76 = v */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 77 = w */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0CC3,  /* 0000110011000011  ....**..**....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 78 = x */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 79 = y */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7A = z */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x3FC3,  /* 0011111111000011  ..********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7B = { */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7C = | */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7D = } */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7E = ~ */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 7F =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFCFF,  /* 1111110011111111  ******..******** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 80 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 81 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 82 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 83 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 84 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 85 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 86 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 87 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 88 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 89 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8A =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8B =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8C =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8D =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8E =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 8F =   */
  {
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 90 =   */
  {
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 91 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xC000,  /* 1100000000000000  **.............. */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C00,  /* 0000110000000000  ....**.......... */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 92 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 93 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 94 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 95 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 96 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 97 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 98 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 99 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9A =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9B =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9C =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x00FF,  /* 0000000011111111  ........******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9D =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9E =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0x000F,  /* 0000000000001111  ............**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x000F,  /* 0000000000001111  ............**** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character 9F =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF03,  /* 1111111100000011  ********......** */
    0xFC30,  /* 1111110000110000  ******....**.... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A0 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A1 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A2 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A3 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A4 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x300F,  /* 0011000000001111  ..**........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A5 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x03C3,  /* 0000001111000011  ......****....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A6 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A7 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A8 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character A9 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AA =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AB =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AC =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC303,  /* 1100001100000011  **....**......** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AD =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AE =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character AF =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character B0 =   */
  {
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
    0xFCFC,  /* 1111110011111100  ******..******.. */
    0xCFCF,  /* 1100111111001111  **..******..**** */
  },
  /* character B1 =   */
  {
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
    0x3333,  /* 0011001100110011  ..**..**..**..** */
    0xCCCC,  /* 1100110011001100  **..**..**..**.. */
  },
  /* character B2 =   */
  {
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
    0x0C0C,  /* 0000110000001100  ....**......**.. */
    0xC0C0,  /* 1100000011000000  **......**...... */
  },
  /* character B3 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character B4 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character B5 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character B6 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character B7 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character B8 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character B9 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character BA =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character BB =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character BC =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character BD =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character BE =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character BF =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character C0 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character C1 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character C2 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character C3 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character C4 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character C5 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0000,  /* 0000000000000000  ................ */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character C6 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character C7 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C0,  /* 1111000011000000  ****....**...... */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character C8 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C0,  /* 1111000011000000  ****....**...... */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF000,  /* 1111000000000000  ****............ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character C9 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF000,  /* 1111000000000000  ****............ */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0C0,  /* 1111000011000000  ****....**...... */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character CA =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x00C0,  /* 0000000011000000  ........**...... */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character CB =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0x00C0,  /* 0000000011000000  ........**...... */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character CC =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C0,  /* 1111000011000000  ****....**...... */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xF0C0,  /* 1111000011000000  ****....**...... */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character CD =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character CE =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x00C0,  /* 0000000011000000  ........**...... */
    0xFFFF,  /* 1111111111111111  **************** */
    0x00C0,  /* 0000000011000000  ........**...... */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character CF =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character D0 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character D1 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character D2 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character D3 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF000,  /* 1111000000000000  ****............ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character D4 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character D5 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character D6 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF000,  /* 1111000000000000  ****............ */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character D7 =   */
  {
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0x0000,  /* 0000000000000000  ................ */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
    0xF0C3,  /* 1111000011000011  ****....**....** */
  },
  /* character D8 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0000,  /* 0000000000000000  ................ */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0000,  /* 0000000000000000  ................ */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character D9 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character DA =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character DB =   */
  {
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
  },
  /* character DC =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
  },
  /* character DD =   */
  {
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
    0x00FF,  /* 0000000011111111  ........******** */
  },
  /* character DE =   */
  {
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
    0xFF00,  /* 1111111100000000  ********........ */
  },
  /* character DF =   */
  {
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0x0000,  /* 0000000000000000  ................ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E0 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E1 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x000F,  /* 0000000000001111  ............**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E2 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E3 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E4 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3F3,  /* 1100001111110011  **....******..** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E5 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E6 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0x3FFF,  /* 0011111111111111  ..************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E7 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E8 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character E9 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0003,  /* 0000000000000011  ..............** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character EA =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0x0303,  /* 0000001100000011  ......**......** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character EB =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF003,  /* 1111000000000011  ****..........** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF00F,  /* 1111000000001111  ****........**** */
    0xC3C3,  /* 1100001111000011  **....****....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character EC =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0x0C30,  /* 0000110000110000  ....**....**.... */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character ED =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFF3,  /* 1111111111110011  ************..** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0F03,  /* 0000111100000011  ....****......** */
    0x0C03,  /* 0000110000000011  ....**........** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0x00C3,  /* 0000000011000011  ........**....** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x0FFF,  /* 0000111111111111  ....************ */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character EE =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC0F,  /* 1111110000001111  ******......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character EF =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC00F,  /* 1100000000001111  **..........**** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0x0FC3,  /* 0000111111000011  ....******....** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F0 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0003,  /* 0000000000000011  ..............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F1 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xC003,  /* 1100000000000011  **............** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F2 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFC3,  /* 1111111111000011  **********....** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F3 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F4 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFF0F,  /* 1111111100001111  ********....**** */
    0xFC03,  /* 1111110000000011  ******........** */
    0xFC33,  /* 1111110000110011  ******....**..** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
  },
  /* character F5 =   */
  {
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xCC3F,  /* 1100110000111111  **..**....****** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F6 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F7 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0C3,  /* 1100000011000011  **......**....** */
    0x0C0F,  /* 0000110000001111  ....**......**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F8 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC03F,  /* 1100000000111111  **........****** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0x0F0F,  /* 0000111100001111  ....****....**** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character F9 =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FA =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FB =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFC00,  /* 1111110000000000  ******.......... */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC03F,  /* 1100000000111111  **........****** */
    0xF03F,  /* 1111000000111111  ****......****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FC =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xC30F,  /* 1100001100001111  **....**....**** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FD =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC0FF,  /* 1100000011111111  **......******** */
    0x0C3F,  /* 0000110000111111  ....**....****** */
    0xFC3F,  /* 1111110000111111  ******....****** */
    0xF0FF,  /* 1111000011111111  ****....******** */
    0xC3FF,  /* 1100001111111111  **....********** */
    0x003F,  /* 0000000000111111  ..........****** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FE =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xC003,  /* 1100000000000011  **............** */
    0xC003,  /* 1100000000000011  **............** */
    0xC003,  /* 1100000000000011  **............** */
    0xC003,  /* 1100000000000011  **............** */
    0xC003,  /* 1100000000000011  **............** */
    0xC003,  /* 1100000000000011  **............** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
  /* character FF =   */
  {
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
    0xFFFF,  /* 1111111111111111  **************** */
  },
};

static unsigned char
*convert( unsigned char *Src,
			  unsigned char *Dest,
			  int Row,
			  int Count)
{
	int SourceChar;
	unsigned short Pattern;
	unsigned char *Save = Src;

	while (Count--)
	{
		SourceChar = *Src++;
		Pattern = CharacterSet [SourceChar & 0xff] [Row];
		*Dest++ = (unsigned char)((Pattern >> 8) & 0xff);
		*Dest++ = (unsigned char)(Pattern & 0xff);
	}

	return Save;
}

int
main(int argc, char **argv)
{
  /* FILE pointers for the text input and the     */
  /* converted binary output                      */
  FILE *input;
  FILE *output;

  /* a buffer to hold the text read from the file  */
  /* the size of the buffer guarantees that it can */
  /* receive CHARS_PER_ROW characters from fgets() */
  /* while having room for a '\n' and a '\0'       */
  char textbuff [CHARS_PER_ROW + 2];

  /* a buffer to hold the converted binary output  */
  /* one scan line at a time                       */
  unsigned char binary [OCTETS_PER_ROW];

  /* the character pointer below is used to check  */
  /* the text read by fgets() for a '\n', and to   */
  /* remove it if present                          */
  char *find_new_line;

  int row;    /* counts the scan line rows         */
  int count;  /* counts the lines converted        */

  /* check number of command line arguments and    */
  /* open the source and destination files, clean  */
  /* up and exit if an error occurs                */
  if (argc < 3)
  {
    fputs("usage: t2b text-file binary-file\n", stderr);
    return EXIT_FAILURE;
  }
  else if (NULL == (input = fopen(argv[1], "r")))
  {
    fprintf(stderr, "t2b: can't open %s", argv[1]);
    return EXIT_FAILURE;
  }
  else if (NULL == (output = fopen(argv[2], "wb")))
  {
    fprintf(stderr, "t2b: can't make %s", argv[2]);
    fclose(input);
    return EXIT_FAILURE;
  }

  /* display successful file open                  */
  fprintf(stderr, "t2b processing\n",
                  "reading text input %s\n",
                  "making binary output %s\n",
                  argv[1], argv[2]);

  /* 64 text lines converted at 16 scan rows per   */
  /* text line equals 1024 scan rows, which equals */
  /* approximately 269 millimeters                 */
  for (count = 0; count < LINES_PER_PAGE; ++count)
  {
    /* fill the input buffer with '\0' characters  */
    /* before each input in case the new line has  */
    /* fewer characters than the line before it    */
    memset(textbuff, '\0', sizeof(textbuff));

    if (NULL == fgets(textbuff, sizeof textbuff, input))
    {
      /* if the file has fewer than the maximum    */
      /* number of lines                           */
      fclose(input);
      fclose(output);
      break;
    }

    /* fgets() leaves the '\n' in the input buffer */
  	/* if it found one, so check for it and remove */
  	/* if present                                  */
    find_new_line = strchr(textbuff, '\n');
    if (find_new_line != NULL)
    {
      *find_new_line = '\0';
    }

    /* now convert to binary scan patterns         */
    for (row = 0; row < SCAN_ROWS_PER_CHAR; ++row)
    {
      convert((unsigned char *)textbuff,
               binary, row, CHARS_PER_ROW);

      /* write the binary to output file           */
      if ((fwrite(binary, 1, OCTETS_PER_ROW, output))
          != OCTETS_PER_ROW)
      {
        fprintf(stderr, "t2b: error writing %s\n", argv[2]);
        fclose(input);
        fclose(output);
        return EXIT_FAILURE;
      }
    }  /* end of ROWS_PER_CHAR repetitions         */
  }  /* end of TEXT_LINES_PER_PAGE or end of input */

  /* display the number of lines converted         */
  printf("Finished converting %d lines\n", count);
  return EXIT_SUCCESS;
}


